
import sys
import os
import logging


class Config:
    """
    Config class provides a way to create and manage a configuration object in Python.
    This class uses the __setattr__, __getattr__, keys, __getitem__, __setitem__, get, 
    and update methods to enable easy access and modification of configuration values.

    Overall, the Config class provides a convenient way to manage configuration values
    in Python with an easy-to-use API.
    """

    def __init__(self):
        self.__docs__ = {}
        self.__fromparam__ = False
        if "__docs__" in self.__docs__:
            del self.__docs__["__docs__"]
        if "__fromparam__" in self.__docs__:
            del self.__docs__["__fromparam__"]

    def __setattr__(self, key: str, value: any):
        """
        Set the attribute `value` to the given `key` in the `__dict__` dictionary of the `Config` object.

        Parameters
        ----------
        key : str
            The key of the attribute.
        value : Any
            The value to be set to the attribute.

        Returns
        -------
        None
        """
        self.__dict__[key] = value

    def __getattr__(self, key: str):
        """
        Retrieve the attribute value for the given `key` from the `__dict__` dictionary of the `Config` object.

        Parameters
        ----------
        key : str
            The key of the attribute.

        Returns
        -------
        Any
            The value of the attribute with the given `key`.
        """
        value = self.__dict__.get(key, None)
        if value is not None:
            return value
        else:
            self.__dict__[key] = Config()
            return self.__dict__[key]

    def keys(self):
        _keys = list(self.__dict__.keys())
        _keys.remove("__docs__")
        _keys.remove("__fromparam__")
        return _keys

    def __getitem__(self, key: str):
        return self.__dict__[key]

    def __setitem__(self, key: str, value: any):
        self.__dict__[key] = value

    def get(self, key: str) -> any:
        return self.__dict__.get(key, None)

    def _update(self, preconfig, config):
        for key in config.keys():
            if isinstance(config[key], Config):
                if key not in preconfig.keys():
                    preconfig[key] = config[key]
                preconfig[key] = self._update(preconfig[key], config[key])
            else:
                preconfig[key] = config[key]
                if key in config.__docs__:
                    preconfig.__docs__ = config.__docs__[key]
        return preconfig

    def update(self, config):
        self = self._update(self, config)

    def __iadd__(self, other):
        name = list(other.keys())[0]
        value = other[name]
        if name in other.__docs__:
            doc = other.__docs__[name]
            self.__docs__[name] = doc
        self[name] = value
        self.__fromparam__ = False
        if "__fromparam__" in self.__docs__:
            del self.__docs__["__fromparam__"]
        return self

    def fromparam(self):
        return self.__fromparam__

    def doc(self, doc):
        """
        Used when Param(db="Colav").doc("MongoDB database name") is called,
        Param only has one key.
        """
        if self.fromparam():
            name = list(self.keys())[0]
            self.__docs__[name] = doc
            return self
        else:
            print("ERROR: this method only can be call from class Param",
                  file=sys.stderr)
            sys.exit(1)

    def set_doc(self, var: str, doc: str):
        self.__docs__[var] = doc

    def _dict(self, config, root=True) -> dict:
        """
        This method creates a dictionary from the config object
        """
        info = {}
        if root:
            cnf = Config()
            cnf.config = config
            config = cnf

        for key in config.keys():
            if isinstance(config[key], Config):
                info[key] = self._dict(config[key], False)
            else:
                value = config[key]
                doc = ""
                if key in config.__docs__:
                    doc = config.__docs__[key]
                info[key] = {'value': value, 'doc': doc}

        return info

    def dict(self):
        return self._dict(self)


class Param:
    def __new__(cls, **kwargs):
        if len(kwargs) == 0:
            print(
                "ERROR: Param can not be empty, at least one param have to be provided ex: Param(var='test')")
            sys.exit(1)

        if len(kwargs) > 2:
            print("ERROR: A maximum of two parameters can be passed, the parameter and the doc for the parameter ex: Param(db='test', doc='databaset name')")
            sys.exit(1)

        name = list(kwargs.keys())[0]
        doc = None
        if len(kwargs) == 2:
            if "doc" in kwargs:
                doc = kwargs["doc"]
                del kwargs["doc"]
            else:
                print(f"ERROR: in Parameter {name}, doc parameter not provide." + os.linesep +
                      "       Two parameters can be provided but the second one have to be 'doc' ex: Param(db='test', doc='databaset name')")
                sys.exit(1)

        name = list(kwargs.keys())[0]
        value = kwargs[name]
        config = Config()
        config.__fromparam__ = True
        config[name] = value
        if doc is not None:
            config.__docs__[name] = doc
        return config


class ConfigGenerator:
    config = Config()

    config += Param(host="localhost", doc="Hostname or ip for flask server.")

    config += Param(port=8080,
                    doc="Port for flask server.")

    config += Param(log_file="hunabku.log",
                    doc="Name for logging file. "
                        "This is only enable in logging level different to DEBUG.")

    config += Param(use_reloader=True,
                    doc="Flask allows to reload the endpoint if something is changed in the code.\n"
                        "Set this False to void reload code.")

    config += Param(info_level=logging.DEBUG,
                    doc="The logging level, default DEBUG, set it to INFO for production.")

    config += Param(apikey=os.environ["HUNABKU_APIKEY"] if "HUNABKU_APIKEY" in os.environ else "colavudea",
                    doc="Apikey for authentication."
                    )

    config += Param(plugin_prefix="hunabku",
                    doc="Hunabku search the plugins using the prefix hunabku,"
                        "but if you want to personalize your own server you can change the prefix"
                    )

    config.apidoc += Param(apidoc_dir='hunabku_website',
                           doc="apidocs output directoy"
                           )

    def generate_config(self, output_file, hunabku, overwrite):
        if len(hunabku.plugins) == 0:
            hunabku.load_plugins(verbose=False)

        output = "from hunabku.Config import Config"+os.linesep
        output += "config = Config()"+os.linesep*2
        output += "######################################" + os.linesep
        output += "# Hunabku Server config options below" + os.linesep
        output += "######################################" + os.linesep

        config_dict = self.parse_config(self.config)
        config_dict = self.parse_paths(config_dict)
        output += self.generate_output(config_dict)

        self.config = Config()

        # appening plugins documentatiom to current object
        for plugin in hunabku.plugins:
            if plugin["package"] not in self.config.keys():
                self.config[plugin["package"]] = Config()

            if plugin["mod_name"] not in self.config[plugin["package"]].keys():
                self.config[plugin["package"]][plugin["mod_name"]] = Config()

            self.config[plugin["package"]][plugin["mod_name"]
                                           ][plugin["class_name"]] = plugin['class'].config
        config_dict = self.parse_config(self.config)
        config_dict = self.parse_paths(config_dict)

        output += "##################################################" + os.linesep
        output += "# Hunabku Plugins config options below" + os.linesep
        output += "##################################################" + os.linesep
        output += "# The structure for config plugins are:" + os.linesep
        output += "# config.hunabku_plugin.File.Class.option = value" + os.linesep
        output += "##################################################" + os.linesep
        output += self.generate_output(config_dict)

        if overwrite:
            with open(output_file, "w") as f:
                f.write(output)
                f.close()
            return True
        else:
            if os.path.exists(output_file):
                return False
            else:
                with open(output_file, "w") as f:
                    f.write(output)
                    f.close()
                return True

    def generate_output(self, config_dict: dict):
        output = ""
        for key in config_dict.keys():
            last = key.split(".")[-1]
            output += f"# {last}"+os.linesep
            doc = config_dict[key]["doc"]
            value = config_dict[key]["value"]

            comments = doc.split(os.linesep)
            for comment in comments:
                output += f"#{comment}"+os.linesep

            if isinstance(value, str):
                output += f'{key} = "{value}"'+os.linesep*2
            else:
                output += f'{key} = {value}'+os.linesep*2
        return output

    def parse_config(self, config: Config, root=True) -> dict:
        """
        This method creates a dictionary from the config object
        """
        info = {}
        if root:
            cnf = Config()
            cnf.config = config
            config = cnf

        for key in config.keys():
            if isinstance(config[key], Config):
                if root:
                    info[key] = self.parse_config(config[key], False)
                else:
                    info[f".{key}"] = self.parse_config(config[key], False)
            else:
                value = config[key]
                doc = ""
                if key in config.__docs__:
                    doc = config.__docs__[key]
                info[key] = {'value': value, 'doc': doc}

        return info

    def parse_paths(self, config: dict, prefix="", root=True) -> dict:

        paths = {}

        for key, payload in config.items():
            if key == "config" and root:
                prefix = "config" + prefix
                _paths = self.parse_paths(config[key], prefix, False)
                paths.update(_paths)

            elif key.startswith("."):
                lprefix = prefix + key
                _paths = self.parse_paths(config[key], lprefix, False)
                paths.update(_paths)

            else:
                lprefix = prefix + '.' + key
                paths[lprefix] = payload

        return paths



from hunabku.HunabkuBase import HunabkuPluginBase, endpoint
from hunabku.Config import Config, Param


class Hello(HunabkuPluginBase):
    config = Config()
    config += Param(myvar="myvalue", doc="this is an example var")
    config.subcategory += Param(port=8080).doc("this is other way to set doc")
    config.subcategory += Param(host="localhost")  # may you want a var without doc

    def __init__(self, hunabku):
        super().__init__(hunabku)

    @endpoint('/hello', methods=['GET'])
    def hello(self):
        """
        @api {get} /hello/:id Hello
        @apiName Hello
        @apiGroup Template

        @apiParam {Number} id Users unique ID.

        @apiSuccess {String} firstname Firstname of the User.
        @apiSuccess {String} lastname  Lastname of the User.
        """
        if self.valid_apikey():
            response = self.app.response_class(
                response=self.json.dumps({'hello': 'world'}),
                status=200,
                mimetype='application/json'
            )
            return response
        else:
            return self.apikey_error()

    @endpoint('/config', methods=['GET'])
    def config_end(self):
        """
        @api {get} /config/ Config
        @apiDescription returns the plugin config, this endpoint doesn´t require apikey
        @apiName Config
        @apiGroup Template

        @apiSuccess {String} firstname Firstname of the User.
        @apiSuccess {String} lastname  Lastname of the User.
        """
        response = self.app.response_class(
            response=self.json.dumps(self.config.dict()),
            status=200,
            mimetype='application/json'
        )
        return response




from flask import (
    request
)
from functools import wraps
from hunabku.Config import Config
import inspect
import os
import sys
import json


class Globals:
    endpoints = {}
    verbose = True


def set_verbose(status):
    Globals.verbose = status


def endpoint(path, methods):
    """
    Specialized decorator to use in the methods of the class that inherit from  HunabkuPluginBase
    this decorator allows to register the path and methods [GET,POST,DELETE,PUT]
    in the flask app.

    example:
    class Hello(HunabkuPluginBase):
    def __init__(self,hunabku):
        super().__init__(hunabku)

    @endpoint('/hello',methods=['GET'])
    def hello(self):
        if self.valid_apikey():
            response = self.app.response_class(
                response=self.json.dumps({'hello':'world'}),
                status=200,
                mimetype='application/json'
            )
            return response
        else:
            return self.send_apikey_error()

    """
    def wrapper(func):
        current_frame = inspect.currentframe()

        caller_frame = inspect.getouterframes(current_frame, 2)[1]
        filename = caller_frame.filename
        package_name = caller_frame.filename.split(
            "endpoints")[0].split(os.sep)[-2]
        class_name, func_name = func.__qualname__.split('.')

        if Globals.verbose:
            print(
                f'------ Adding endpoint {path} with HTTP(S) methods {str(methods)} from class = {class_name} class method = {func_name}')
        if package_name not in Globals.endpoints:
            Globals.endpoints[package_name] = []
        Globals.endpoints[package_name].append(
            {'path': path, 'methods': methods, 'func_name': func_name, 'class_name': class_name, 'file': filename})

        @wraps(func)
        def _impl(self, *method_args, **method_kwargs):
            response = func(self, *method_args, **method_kwargs)
            return response
        # WARNING: this is required to avoid overwrite methods in the class
        _impl.__name__ = func.__qualname__
        return _impl
    return wrapper


class HunabkuPluginBase(object):
    config = Config()

    def __init__(self, hunabku):
        """
        Base class to handle the plugins.
        Allows to have access to the MondoDB object, custom json methods
        with our encoders, utility functions to check apikeys and send default messages
        in case of error.

        This class allows to register the endpoints setting a decorator in the method
        that is going to hanlde the endpoint.

        """
        self.global_config = hunabku.config
        self.app = hunabku.app
        self.request = request
        self.json = json
        self.logger = hunabku.logger
        self.hunabku = hunabku

    def apikey_error(self):
        """
        return defualt apikey error
        """
        response = self.app.response_class(
            response=self.json.dumps(
                {'msg': 'The HTTP 401 Unauthorized invalid authentication apikey for the target resource.'}),
            status=401,
            mimetype='application/json'
        )
        return response

    def badrequest_error(self):
        """
        return defualt bad request error
        """
        data = {"error": "Bad Request",
                "message": "Invalid parameters passed. Please fix your request with valid parameters."}
        response = self.app.response_class(response=self.json.dumps(data),
                                           status=400,
                                           mimetype='application/json')
        return response

    def valid_apikey(self):
        if self.request.method == 'POST':
            apikey = self.request.form.get('apikey')
        else:
            apikey = self.request.args.get('apikey')
        if self.global_config["apikey"] == apikey:
            return True
        else:
            return False

    def register_endpoints(self):
        """
        Method to register all the endpoints in flask's app
        """
        filename = inspect.getfile(self.__class__)
        class_name = type(self).__name__
        if self.has_valid_endpoints():
            for endpoint_data in Globals.endpoints[self._get_package_name()]:
                if endpoint_data['file'] == filename and endpoint_data['class_name'] == class_name:
                    path = endpoint_data['path']
                    func_name = endpoint_data['func_name']
                    methods = endpoint_data['methods']
                    func = getattr(self, func_name)
                    self.app.add_url_rule(
                        path, view_func=func, methods=methods)
        else:
            sys.exit(1)

    def valid_parameters(self, params):
        """
        Method to check is the parameters passed to the endpoint are valid,
        if unkown parameter is passed, a bad request should be returned.
        """
        if self.request.method == 'POST':
            args = self.request.form
        else:
            args = self.request.args

        for rarg in args:
            if rarg not in params:
                return False
        return True

    @classmethod
    def get_global_endpoints(cls):
        """
        Method to return the global dictionary with all
        the registers  loaded
        """
        return Globals.endpoints

    def has_valid_endpoints(self):
        """
        This method checks before to load the plugin if any paths in the endpoint is repeated.
        this platform does not allows overwrite endpoint paths.
        """
        package_name = self._get_package_name()
        class_name = type(self).__name__
        plugins = list(Globals.endpoints.keys())
        endpoints = []
        for register in Globals.endpoints[package_name]:
            endpoints.append(register)
        # checking if there are endpoints repeated in the same package
        for register in Globals.endpoints[package_name]:
            for endpoint in endpoints:
                if endpoint != register:
                    if endpoint["path"] == register["path"]:
                        print(
                            f"ERROR: can't not load plugin, package {package_name} class {class_name} class_method {endpoint['func_name']} file {endpoint['file']} "
                            f"because the path {endpoint['path']} is already loaded in plugin: package {package_name} class {register['class_name']} "
                            f"class_method {register['func_name']} file {register['file']}")
                        return False
        plugins.remove(package_name)
        # checking if there are endpoints repeated in other packages
        for endpoint in endpoints:
            for plugin in plugins:
                for register in Globals.endpoints[plugin]:
                    if endpoint["path"] == register['path']:
                        print(
                            f"ERROR: can't not load plugin, package {package_name} class {class_name} class_method {endpoint['func_name']} file {endpoint['file']} "
                            f"because the path {endpoint['path']} is already loaded in plugin: package {plugin} class {register['class_name']} "
                            f"class_method {register['func_name']} file {register['file']}")
                        return False
        return True

    def _get_package_name(self):
        filename = inspect.getfile(self.__class__)
        package_name = filename.split(
            "endpoints")[0].split(os.sep)[-2]
        return package_name



from hunabku import templates
from shutil import copytree, move, rmtree
import pathlib
import os


class PluginGenerator:
    """
    Class to generate a hunabku plugin from the template.
    The generated plugin is a python package ready to install with pip and it can be deployed in pipy.
    """

    def __init__(self, name: str, prefix: str = "HunabKu"):
        """
        Initializes the class with a given name and prefix. The default prefix is "HunabKu".

        Parameters:
        ____________
        name:str
            plugin name for the new project
        prefix:str
            prefix for the generated project, default is Hunabku
        """
        self.name = name
        self.tname = "template"
        self.prefix = prefix
        self.template_path = os.path.join(
            str(pathlib.Path(templates.__file__).parent.absolute()), "plugin")
        self.template_path = os.path.join(
            self.template_path, "HunabKu_template")

    def replace(self, filename, old, new):
        """
        Replaces all occurrences of old with new in a given file.

        Parameters:
        ____________
        filename:str
            filename to repalce the string
        old:str
            string to be replaced
        new:str
            new string to write
        """
        content = ""
        with open(filename, "r") as file:
            content = file.read()
            content = content.replace(old, new)
            file.close()
        with open(filename, "w") as file:
            file.write(str(content))
            file.close()

    def generate(self, path: str = os.getcwd()):
        """
        Generates the plugin in a given path. Copies the template to the output path,
        renames the package folder, and replaces the template name with the plugin name in the README.md,
        setup.py, and MANIFEST.in files.

        Parameters:
        ____________
        path:str
            path to save the plugin project.
        """
        output_path = os.path.join(os.getcwd(), f"{self.prefix}_{self.name}")
        readme = os.path.join(output_path, "README.md")
        setup = os.path.join(output_path, "setup.py")
        manifest = os.path.join(output_path, "MANIFEST.in")
        package_folder_old = os.path.join(output_path, f"hunabku_template")
        package_folder_new = os.path.join(
            output_path, f"{self.prefix}_{self.name}".lower())
        package_folder_endpoints = os.path.join(
            package_folder_new, "endpoints")

        copytree(self.template_path, output_path)
        move(package_folder_old, package_folder_new)

        rmtree(os.path.join(output_path, "__pycache__"), ignore_errors=True)
        rmtree(os.path.join(package_folder_new,
               "__pycache__"), ignore_errors=True)
        rmtree(os.path.join(package_folder_endpoints,
               "__pycache__"), ignore_errors=True)

        self.replace(readme, self.tname, self.name)
        self.replace(setup, self.tname, self.name)
        self.replace(manifest, self.tname, self.name)



from flask import (
    Flask
)

import logging

import os
from hunabku.HunabkuBase import HunabkuPluginBase
from hunabku.Config import ConfigGenerator, Config
from hunabku._version import get_version
from shutil import rmtree
from distutils.dir_util import copy_tree
import subprocess
import inspect
import glob
import time
import pathlib
import sys
import importlib
import json

import pkgutil


class Hunabku:
    """
    Class to serve papers information store in a mongodb database through an API using flask.

    example:
    http://0.0.0.0:5000/data/redalyc?init=1&end&apikey=pl0ok9ij8uh7yg
    """
    config = ConfigGenerator.config

    def __init__(self, config: Config):
        """
        Contructor to initialize configuration options.

        Args:
            apikey: apikey to access the data
            ip (str): ip to start the server
            port (int): port for the server
            info_level (logging.DEBUG/INFO etc..): enable/disable debug mode with extra messages output.
        """
        self.config.update(config)
        self.plugin_prefix = "hunabku"
        self.apidoc_dir = self.config.apidoc.apidoc_dir
        self.apidoc_static_dir = self.apidoc_dir + '/static'
        self.apidoc_output_dir = self.apidoc_dir + '/static/apidoc'
        self.apidoc_templates_dir = self.apidoc_dir + '/templates'
        self.apidoc_config_dir = self.apidoc_dir + '/config'
        self.apidoc_config_data = {}
        self.apidoc_config_data['url'] = 'http://' + \
            config.host + ':' + str(config.port)
        self.apidoc_config_data['sampleUrl'] = 'http://' + \
            config.host + ':' + str(config.port)
        self.apidoc_config_data['header'] = {}
        self.apidoc_config_data['header']['filename'] = self.apidoc_config_dir + \
            '/apidoc-header.md'
        self.apidoc_config_data['version'] = get_version()
        self.pkg_config_dir = str(
            pathlib.Path(__file__).parent.absolute()) + '/config/'
        self.pkg_templates_dir = str(
            pathlib.Path(__file__).parent.absolute()) + '/templates/'
        self.plugins = []
        self.logger = logging.getLogger(__name__)
        self.set_info_level(config["info_level"])
        self.app = Flask(
            'hanubku',
            static_folder=self.apidoc_static_dir,
            static_url_path='/',
            template_folder=self.apidoc_templates_dir)

    def apidoc_setup(self):
        """
        creates an ApiDoc folder to dump configuration and documentation of the APIs
        """
        try:
            os.makedirs(self.apidoc_dir, exist_ok=True)
            print(" * ApidDoc Directory ", self.apidoc_dir, " created ")
        except FileExistsError:
            # Is this is happening then restart the microservices in the folder
            print(
                " * Warning! ApidDoc Directory ",
                self.apidoc_dir,
                " already exists")

        try:
            os.mkdir(self.apidoc_static_dir)
            print(
                " * ApiDoc Static Directory ",
                self.apidoc_static_dir,
                " created ")
        except FileExistsError:
            # Is this is happening then restart the microservices in the folder
            print(
                " * Warning! ApiDoc Static Directory ",
                self.apidoc_static_dir,
                " already exists")

        try:
            os.mkdir(self.apidoc_output_dir)
            print(" * ApiDoc Static Output Directory ",
                  self.apidoc_output_dir, " created ")
        except FileExistsError:
            # Is this is happening then restart the microservices in the folder
            print(
                " * Warning! ApiDoc Static Output Directory ",
                self.apidoc_output_dir,
                " already exists")

        try:
            os.mkdir(self.apidoc_templates_dir)
            print(
                " * ApiDoc Templates Directory ",
                self.apidoc_templates_dir,
                " created ")
        except FileExistsError:
            # Is this is happening then restart the microservices in the folder
            print(
                " * Warning! ApiDoc Templates Directory ",
                self.apidoc_templates_dir,
                " already exists")

        try:
            os.mkdir(self.apidoc_config_dir)
            print(
                " * ApiDoc Config Directory ",
                self.apidoc_static_dir,
                " created ")
        except FileExistsError:
            # Is this is happening then restart the microservices in the folder
            print(
                " * Warning! ApiDoc Static Directory ",
                self.apidoc_static_dir,
                " already exists")

        copy_tree(self.pkg_config_dir, self.apidoc_config_dir)
        apidoc_config_data = {}
        with open(self.apidoc_config_dir + '/apidoc.json') as json_file:
            apidoc_config_data = json.load(json_file)

        apidoc_config_data.update(self.apidoc_config_data)
        with open(self.apidoc_config_dir + '/apidoc.json', 'w') as json_file:
            json.dump(apidoc_config_data, json_file)

        copy_tree(self.pkg_templates_dir, self.apidoc_templates_dir)

    def set_info_level(self, info_level):
        """
        Information level for debug or verbosity of the application (https://docs.python.org/3.1/library/logging.html)
        """
        if info_level != logging.DEBUG:
            logging.basicConfig(
                filename=self.config["log_file"], level=info_level)
        self.config["info_level"] = info_level

    def load_plugins(self, verbose=True):
        """
        This method return the plugins found in the folder plugins.
        """
        if verbose:
            self.logger.warning('-----------------------')
            self.logger.warning('------ Loading Plugins:')
        discovered_plugins = {
            name: importlib.import_module(name)
            for finder, name, ispkg
            in pkgutil.iter_modules()
            if name.startswith(self.plugin_prefix + '_')
        }
        for discovered_plugin in discovered_plugins:
            for path in glob.glob(
                    str(discovered_plugins[discovered_plugin].__path__[0]) + "/endpoints/*.py"):
                mname = path.split(os.path.sep)[-1].replace('.py', '')
                if verbose:
                    self.logger.warning(
                        f'------ Loading plugin module from package {discovered_plugin} and module {mname}.py :')
                spec = importlib.util.spec_from_file_location(mname, path)
                module = spec.loader.load_module()
                for cname, plugin_class in inspect.getmembers(module):
                    if inspect.isclass(plugin_class) and issubclass(plugin_class, HunabkuPluginBase) and plugin_class is not HunabkuPluginBase:
                        if verbose:
                            self.logger.warning(
                                f'------ Registering plugin class: {mname}.{cname}')
                        instance = plugin_class(self)
                        instance.register_endpoints()
                        plugin = {}
                        plugin['package'] = discovered_plugin
                        plugin['mod_name'] = mname
                        plugin['class'] = plugin_class
                        plugin['class_name'] = cname
                        plugin['name'] = f"{discovered_plugin}.{mname}.{cname}"
                        plugin['path'] = path
                        plugin['spec'] = spec
                        plugin['instance'] = instance
                        if discovered_plugin in self.config.keys():
                            if mname in self.config[discovered_plugin].keys():
                                if cname in self.config[discovered_plugin][mname].keys():
                                    instance.config.update(
                                        self.config[discovered_plugin][mname][cname])
                        self.plugins.append(plugin)
                        if verbose:
                            self.logger.warning(
                                f'------ Registered plugin class: {mname}.{cname}  DONE')

    def check_apidoc_syntax(self, plugin_file):
        """
        Allows to check in the syntaxis in the docstring comment is right
        for apidoc  files generation.
        The the syntax is wrong, the Hunabku server can not start.
        """
        args = ['apidoc', '-c', self.apidoc_config_dir + os.path.sep + "apidoc.json", '-i',
                str(pathlib.Path(plugin_file).parent.absolute()),
                '--dry-run',
                '-f',
                plugin_file]
        process = subprocess.run(args,
                                 stdout=subprocess.PIPE)
        if process.returncode != 0:
            self.logger.error(
                '------ERROR: parsing docstring for apidocs in plugin ' + plugin_file)
            self.logger.error(
                '             server can not start until apidocs syntax is fixed')
            self.logger.error(process)
            sys.exit(1)

    def generate_doc(self, timeout=1, maxtries=5):
        """
        this method allows to generated apidocs documentation parsing plugin files
        """
        self.logger.warning('-----------------------')
        self.logger.warning('------ Creating documentation')

        rmtree(self.apidoc_static_dir, ignore_errors=True)
        args = ['apidoc', '-c', self.apidoc_config_dir +
                os.path.sep + "apidoc.json"]

        for plugin in self.plugins:
            self.check_apidoc_syntax(plugin['path'])
            args.append('-i')
            args.append(str(pathlib.Path(plugin['path']).parent.absolute()))
            args.append('-f')
            args.append(plugin['path'])
        args.append('-o')
        args.append(self.apidoc_output_dir)
        process = subprocess.Popen(args,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT)
        counter = 0
        while process.poll() is None:
            time.sleep(timeout)
            counter = counter + 1
            if counter == maxtries:
                process.kill()
                break
        self.logger.warning(
            '------ Apidocs at http://{}:{}/apidoc/index.html'.format(self.config.host, self.config.port))

    def start(self):
        """
        Method to start server
        """
        self.app.run(host=self.config.host, port=self.config.port,
                     debug=True, use_reloader=self.config.use_reloader)
